from abc import ABC
from typing import Any, Tuple

import openai
import tiktoken

from llmpdf import db, util

COST_TABLE = {
    "gpt-4": {
        "input": 0.03 / 1000,
        "output": 0.06 / 1000,
    },
    "gpt-3.5-turbo": {
        "input": 0.0015 / 1000,
        "output": 0.002 / 1000,
    },
}


class AINode(ABC):
    def __init__(self, collection_name: str):
        self.settings = util.Settings()
        self.collection_name = collection_name
        self.database = db.Database()
        self.cost = 0.0

    def estimate_tokens(self) -> Tuple[int, int]:
        """Estimate the number of tokens in the input and output."""
        enc = tiktoken.encoding_for_model(self.settings.model)
        num_pages = self.database.count(self.collection_name)
        pages = self.database.list(self.collection_name, limit=num_pages)
        return (
            sum([len(enc.encode(page)) for page in pages]),
            self.settings.max_tokens,
        )

    def calculate_cost(self, response: Any = None) -> float:
        if response:
            input_tokens = response.usage.prompt_tokens
            output_tokens = response.usage.completion_tokens
        else:
            input_tokens, output_tokens = self.estimate_tokens()

        input_cost = COST_TABLE[self.settings.model]["input"] * input_tokens
        output_cost = COST_TABLE[self.settings.model]["output"] * output_tokens
        return input_cost + output_cost


class Summarizer(AINode):
    """Summarizes text."""

    ZERO_SHOT_PROMPT = """
You are a diligent, precise, and concise summarizer. The following text has been selected from a PDF document. Summarize it in one paragraph.
Use only the provided text.

TEXT:
{text}
"""  # noqa: E501

    def zero_shot(self) -> Tuple[str, float]:
        """Summarize the text in one go."""
        text = "".join(self.database.list(self.collection_name))
        prompt = self.ZERO_SHOT_PROMPT.strip().format(text=text)

        openai.api_key = self.settings.openai_api_key
        response = openai.ChatCompletion.create(
            model=self.settings.model,
            max_tokens=self.settings.max_tokens,
            temperature=self.settings.temperature,
            messages=[{"role": "user", "content": prompt}],
        )
        cost = self.calculate_cost(response)
        return response.choices[0].message.content, cost  # type: ignore

    def summarize(self) -> Tuple[str, float]:
        """Summarize the text. Return the text and the cost."""
        input_tokens, _ = self.estimate_tokens()
        if input_tokens + self.settings.max_tokens < 8000:
            return self.zero_shot()

        return "This is a summary.", 0.0


class ChatBot(AINode):
    """Chat with your PDFs."""

    START_PROMPT = """
You are an assistant for question-answering tasks. I will provide you pages of a PDF file that you can use to answer your question.
Do not use prior information you may have about the topic.
If you do not know the answer, say that you do not know.

Use five sentences maximum and keep the answer concise.

Context:
{context}

Question:
{question}

Answer:
"""  # noqa: E501

    def __init__(self, collection_name: str):
        """Initialize the chatbot."""
        super().__init__(collection_name)
        self.memory = []

    def _messages(self, question: str, results: list[db.Result]):
        """Make the messages for the chatbot."""
        # Get context from the database
        context = "\n--\n".join(
            [f"Page {result.page}:\n{result.document}" for result in results]
        )
        prompt = self.START_PROMPT.strip().format(context=context, question=question)

        self.memory.append({"role": "user", "content": prompt})
        return self.memory

    def chat(self, question: str) -> Tuple[str, float]:
        """Chat with the database."""
        results = self.database.search(self.collection_name, question)
        if not results:
            return "I don't have anything in my database about this question.", 0.0

        openai.api_key = self.settings.openai_api_key
        response = openai.ChatCompletion.create(
            model=self.settings.model,
            messages=self._messages(question, results),
            max_tokens=self.settings.max_tokens,
            temperature=self.settings.temperature,
        )

        cost = self.calculate_cost(response)
        self.cost += cost
        return response.choices[0].message.content, cost  # type: ignore



from typing import Generator

import arxiv

from llmpdf.util import Settings

settings = Settings()


class Arxiv:
    """Download papers from arXiv"""

    def __init__(self):
        self.download_dir = settings.base_dir / "arxiv"
        self.download_dir.mkdir(parents=True, exist_ok=True)

    def search(self, query: str) -> Generator[arxiv.Result, None, None]:
        """Search arXiv"""
        self.query = query
        search = arxiv.Search(query=query, max_results=10)
        return search.results()

    def download(self, id: str):
        """Download a single paperfrom arXiv"""
        search = arxiv.Search(id_list=[id])
        result = next(search.results())
        if result:
            return result.download_pdf(dirpath=str(self.download_dir))
        else:
            raise ValueError(f"Could not find paper with id {id}")

    def download_url(self, url: str):
        """Download a single paperfrom arXiv"""
        paper = next(self.search(url))
        if paper:
            print(f"Downloading {paper.title}...")



from pathlib import Path
from typing import List
from urllib.parse import urlparse

import click
from rich.console import Console
from rich.progress import track
from rich.prompt import Confirm, Prompt

from llmpdf import __version__, util
from llmpdf.ai import ChatBot, Summarizer
from llmpdf.db import Database
from llmpdf.download import Arxiv

settings = util.Settings()
console = Console()


def expand_globs(pdfs: list[str]) -> list[str]:
    files = []
    for pdf in pdfs:
        pdf = Path(pdf).expanduser()
        if "*" in pdf.name:
            files.extend(Path.cwd().glob(pdf.name))
        else:
            files.append(pdf)

    files = [str(Path(file).resolve()) for file in files if Path(file).is_file()]

    return files


@click.group()
@click.option("-4", "--gpt-4", is_flag=True, default=False, help="Use GPT-4.")
@click.version_option(__version__)
def main(gpt_4: bool):
    pass


@main.command(name="list")
@click.option(
    "-c",
    "--collection",
    "collection_name",
    required=False,
    type=str,
    help="Name of the collection.",
)
def ls(collection_name: str):
    """Reset the collection."""
    database = Database()
    results = database.list(collection_name)
    console.print(results)


@main.command()
@click.option(
    "-c",
    "--collection",
    "collection_name",
    type=str,
    required=True,
    help="Name of the collection.",
)
def delete(collection_name: str):
    """Reset the collection."""
    if not Confirm.ask(
        f"Are you sure you want to delete {collection_name}?", default=False
    ):
        return

    database = Database()
    database.delete(collection_name)

    console.print(f"Deleted {collection_name}.")


@main.command()
@click.option(
    "-c",
    "--collection",
    "collection_name",
    type=str,
    required=True,
    help="Name of the collection.",
)
@click.option("-n", "--dry-run", is_flag=True, default=False, help="Dry run.")
@click.argument("pdfs", nargs=-1, type=click.Path(exists=True))
def index(
    collection_name: str,
    pdfs: List[str],
    dry_run: bool = False,
):
    """Chat with your PDFs."""

    database = Database()

    pdfs = expand_globs(pdfs)
    console.print(f"Indexing PDFs {pdfs}")

    for pdf in track(pdfs):
        if not dry_run:
            database.index(collection_name, pdf)


@main.command()
@click.option(
    "-c",
    "--collection",
    "collection_name",
    type=str,
    required=True,
    help="Name of the collection.",
)
@click.option("-n", "--dry-run", is_flag=True, default=False, help="Dry run.")
@click.argument("query", nargs=-1, type=str)
def arxiv(
    collection_name: str,
    query: List[str],
    dry_run: bool,
):
    """Download and index a paper from arxiv."""
    query_str = " ".join(query)

    downloader = Arxiv()
    database = Database()

    if query_str.startswith("http"):
        # If query_str is an arxiv url of the type http://arxiv.org/abs/2107.05580v1
        # or of the type http://arxiv.org/abs/quant-ph/0201082v1
        # extract an id which is the part of the URL after /abs/
        parsed_url = urlparse(query_str)
        if parsed_url.netloc == "arxiv.org" and parsed_url.path.startswith("/abs/"):
            arxiv_id = parsed_url.path.split("/abs/")[1]

            with console.status(f"Downloading {arxiv_id}..."):
                if not dry_run:
                    file_path = downloader.download(arxiv_id)
                else:
                    file_path = "dummy.pdf"

            with console.status(f"Indexing {file_path}..."):
                if not dry_run:
                    database.index(collection_name, file_path)
        else:
            console.print("Invalid arxiv URL.")
            return
    else:
        with console.status("[bold green]Searching arXiv..."):
            results = list(downloader.search(query_str))
        console.print(
            "\n".join([f"{idx}. {result.title}" for idx, result in enumerate(results)])
        )
        choices = list(str(i) for i in range(len(results)))
        choice = Prompt.ask("Choose a paper", choices=choices)
        result = results[int(choice)]

        with console.status(f"Downloading {result.title}..."):
            if not dry_run:
                file_path = result.download_pdf(dirpath=str(downloader.download_dir))
                console.print(f"Downloaded {file_path}")
            else:
                file_path = "dummy.pdf"

        with console.status(f"Indexing {file_path}..."):
            if not dry_run:
                database.index(collection_name, file_path)


@main.command()
@click.option(
    "-c",
    "--collection",
    "collection_name",
    type=str,
    required=True,
    help="Name of the collection.",
)
def chat(collection_name: str):
    """Chat with your PDFs."""
    console.print(f"Chatting with {collection_name}.")
    console.print("Type 'exit' to exit.")
    console.print("Type 'new' to start a new chat.")
    chatbot = ChatBot(collection_name)
    while True:
        query = Prompt.ask(
            "Question",
        )
        if query == "exit":
            return
        elif query == "new":
            console.print("Starting a new chat.")
            chatbot = ChatBot(collection_name)
            continue
        else:
            with console.status("[bold green] Asking 🤖..."):
                response, cost = chatbot.chat(query)
            console.print(response)
            console.print(f"Cost: ${cost:.2f}")
            console.print(f"Total Cost: ${chatbot.cost:.2f}")


@main.command()
@click.option(
    "-c",
    "--collection",
    "collection_name",
    type=str,
    required=True,
    help="Name of the collection.",
)
def summarize(collection_name: str):
    """Chat with your PDFs."""
    summarizer = Summarizer(collection_name=collection_name)
    if not Confirm.ask(
        f"Summarization will cost {sum(summarizer.estimate_tokens())} tokens and ${summarizer.calculate_cost():.2f}. Continue?",  # noqa: E501
        default=False,
    ):
        return

    # summary = summarizer.summarize()
    with console.status("[bold green]Summarizing..."):
        summary, cost = summarizer.summarize()

    console.print(summary)
    console.print(f"Cost: ${cost:.2f}")


if __name__ == "__main__":
    main()



from typing import List

import chromadb
from chromadb.utils import embedding_functions
from pydantic import BaseModel
from pypdf import PdfReader

from llmpdf import util


class Result(BaseModel):
    """A search result."""

    document: str
    page: str


class Database:
    """Wraps the SQLite and Vector Databases."""

    embedding_model = "text-embedding-ada-002"

    def __init__(self):
        """Initialize the database."""
        settings = util.Settings()
        chroma_dir = settings.chroma_dir
        self.db_client = chromadb.PersistentClient(path=str(chroma_dir))
        self.embedding_fn = embedding_functions.OpenAIEmbeddingFunction(
            api_key=settings.openai_api_key,
            model_name=self.embedding_model,
        )

    def _get(self, collection_name: str):
        """Get the collection."""
        return self.db_client.get_collection(
            collection_name, embedding_function=self.embedding_fn
        )

    def index(self, collection_name: str, pdf: str):
        """Index PDFs."""
        coll = self.db_client.get_or_create_collection(
            collection_name, embedding_function=self.embedding_fn
        )
        reader = PdfReader(pdf)

        for page in reader.pages:
            id = page.page_number
            text = page.extract_text() + " "
            metadata = {"page": str(id)}

            coll.add(documents=text, ids=str(id), metadatas=metadata)

    def search(self, collection_name: str, query: str) -> List[Result]:
        """Search the database."""
        coll = self._get(collection_name)
        res = coll.query(query_texts=query, n_results=3)
        documents = res["documents"][0] if res["documents"] else []
        metadatas = res["metadatas"][0] if res["metadatas"] else []
        results = [
            Result(document=document, page=str(metadata["page"]))
            for (document, metadata) in zip(documents, metadatas)
        ]
        return results

    def delete(self, collection_name: str):
        """Reset the database."""
        try:
            self.db_client.delete_collection(collection_name)
        except ValueError:
            pass

    def list(self, collection_name: str | None = None, limit: int = 5) -> list[str]:
        """List collections or documents in the collection."""
        if collection_name is None:
            collections = self.db_client.list_collections()
            return [c.name for c in collections]
        else:
            coll = self._get(collection_name)
            return coll.peek(limit=limit)["documents"] or []

    def count(self, collection_name: str):
        """Count the number of documents in the collection."""
        coll = self._get(collection_name)
        return coll.count()



from pathlib import Path

from pydantic_settings import BaseSettings, SettingsConfigDict

BASE_DIR = Path("~/.llm-pdf").expanduser()
SETTINGS_FILE = BASE_DIR / "settings.env"
CHROMA_DIR = BASE_DIR / "chroma"

BASE_DIR.mkdir(parents=True, exist_ok=True)
SETTINGS_FILE.touch(exist_ok=True)
CHROMA_DIR.mkdir(parents=True, exist_ok=True)


class Settings(BaseSettings):
    model_config = SettingsConfigDict(env_file=SETTINGS_FILE)

    openai_api_key: str = ""
    model: str = "gpt-4"
    embedding_model: str = "text-embedding-ada-002"
    temperature: float = 0.0
    max_tokens: int = 500

    chroma_dir: Path = CHROMA_DIR
    base_dir: Path = BASE_DIR
